/*
 *  PCI Express Scan 1.0 Alpha
 *  Copyright (R) 2019  Asterix
 *  Released to the Public Domain
 */

#include <stdio.h>

int checkRSDP (void *rsdp_in) {
 unsigned char *i, *rsdp = rsdp_in, a = 0;

 if (!rsdp) {
  return 0;
 }

 for (i = rsdp; i < &rsdp[20]; i++) {
  a += *i;
 }

 return !a;
}

void printRSDP (void *rsdp_in) {
 unsigned char *rsdp = rsdp_in;

 if (checkRSDP (rsdp)) {
  printf ("Signature: %c%c%c%c%c%c%c%c\n", rsdp[0], rsdp[1], rsdp[2]
   , rsdp[3], rsdp[4], rsdp[5], rsdp[6], rsdp[7]);
  printf ("Checksum: 0x%hhx\n", rsdp[8]);
  printf ("OEMID: 0x%x%hx\n", *(unsigned int *) &rsdp[9]
   , *(unsigned short *) &rsdp[13]);
  printf ("Revision: 0x%hhx\n", rsdp[15]);
  printf ("RSDT Address: 0x%p\n", *(void **) &rsdp[16]);
 } else {
  printf ("No valid RSDP found at 0x%p\n", rsdp);
 }
}

void *getRSDP (void) {
 unsigned char *rsdp;

 for (rsdp = (void *) 0; rsdp <= (void *) 0x000fffff; rsdp += 1) {
  if ('R' == rsdp[0] && 'S' == rsdp[1] && 'D' == rsdp[2] && ' ' == rsdp[3]
   && 'P' == rsdp[4] && 'T' == rsdp[5] && 'R' == rsdp[6]
   && ' ' == rsdp[7]) {
   printf ("RSDP found at 0x%p", rsdp);
   if (checkRSDP (rsdp)) {
    printf (" Checksum OK\n");
    return rsdp;
   } else {
    printf (" Checksum not OK\n");
    continue;
   }
  } else {
   continue;
  }
 }
 printf ("RSDP not found\n");
 return 0;
}

int checkRSDT (void *rsdt_in) {
 unsigned char a = 0, i, *rsdt = rsdt_in, l_rsdt;

 if (!rsdt) {
  return 0;
 }
 
 l_rsdt = *(unsigned int *) &rsdt[4];
 for (i = 0; i < l_rsdt; i++) {
  a += rsdt[i];
//  printf ("%hhu\n",a);
 }
 return !a;
}

void *getRSDT (void *rsdp_in) {
 unsigned char a, i, *rsdp = rsdp_in, *rsdt, l_rsdt;
 
 if (checkRSDP (rsdp)) {
  rsdt = *(void **) &rsdp[16];
  l_rsdt = *(unsigned int *) &rsdt[4];
  if (checkRSDT (rsdt)) {
   printf ("Valid RSDT found at 0c%p\n", rsdt);
   return rsdt;
  } else {
   printf ("RSDT invalid\n");
   return 0;
  }
 } else {
  printf ("RSDP invalid\n");
  return 0;
 }
}

void printRSDT (void *rsdt_in) {
 unsigned char a, i, **entry, *table, *rsdt = rsdt_in, l_rsdt;
 
 l_rsdt = *(unsigned int *) &rsdt[4];
 if (checkRSDT (rsdt)) {
  printf ("Signature: %c%c%c%c\n", rsdt[0], rsdt[1], rsdt[2], rsdt[3]);
  printf ("Length: 0x%x\n", *(unsigned int *) &rsdt[4]);
  printf ("Revision: 0x%hhx\n", rsdt[8]);
  printf ("Checksum: 0x%hhx\n", rsdt[9]);
  printf ("OEMID: 0x%x%hx\n", *(unsigned int *) &rsdt[10]
   , *(unsigned short *) &rsdt[14]);
  printf ("OEM Table ID: 0x%x%x\n", *(unsigned int *) &rsdt[16]
   , *(unsigned int *) &rsdt[20]);
  printf ("OEM Revision: 0x%x\n", *(unsigned int *) &rsdt[24]);
  printf ("Creator ID: 0x%x\n", *(unsigned int *) &rsdt[28]);
  printf ("Creator Revisoin: 0x%x\n", *(unsigned int *) &rsdt[32]);
  for (entry = (unsigned char **) &rsdt[36]; entry < rsdt + l_rsdt
   ; entry += 1) {
   table = *entry; 
   printf ("Entry at 0x%p pointing to table %c%c%c%c at 0x%p\n", entry
    , table[0], table[1], table[2], table[3], table);
  }
 } else {
  printf ("RSDT invalid\n");
 }
}

int checkMCFG (void *mcfg_in)  {
 unsigned char *mcfg = mcfg_in;
 unsigned int i, a = 0, l_mcfg = *(unsigned int *) &mcfg[4];

 if (!mcfg) {
  return 0;
 }
 
 for (i = 0; i < l_mcfg; i++) {
  a += mcfg[i];
 }
 
 return !(a & 0xff);
}

void *getMCFG (void *rsdt_in) {
 unsigned char **entry, *table, *rsdt = rsdt_in, l_rsdt;
 
 l_rsdt = *(unsigned int *) &rsdt[4];
 if (checkRSDT (rsdt)) {
  for (entry = (unsigned char **) &rsdt[36]; entry < rsdt + l_rsdt
   ; entry += 1) {
   table = *entry; 
   if ('M' == table[0] && 'C' == table[1] && 'F' == table[2]
    && 'G' == table[3]) {
    printf ("MCFG found\n");
    return table;
   }
  }
 } else {
  printf ("Invalid RSDT given\n");
  return 0;
 }
 printf ("MCFG not found in RSDT\n");
 return 0;
}

void printMCFG (void *mcfg_in) {
 unsigned char a, i, *entry, *table, *mcfg = mcfg_in, l_mcfg;
 
 l_mcfg = *(unsigned int *) &mcfg[4];
 if (checkMCFG (mcfg)) {
  printf ("Signature: %c%c%c%c\n", mcfg[0], mcfg[1], mcfg[2], mcfg[3]);
  printf ("Length: 0x%x\n", *(unsigned int *) &mcfg[4]);
  printf ("Revision: 0x%hhx\n", mcfg[8]);
  printf ("Checksum: 0x%hhx\n", mcfg[9]);
  printf ("OEMID: 0x%x%hx\n", *(unsigned int *) &mcfg[10]
   , *(unsigned short *) &mcfg[14]);
  printf ("OEM Table ID: 0x%x%x\n", *(unsigned int *) &mcfg[16]
   , *(unsigned int *) &mcfg[20]);
  printf ("OEM Revision: 0x%x\n", *(unsigned int *) &mcfg[24]);
  printf ("Creator ID: 0x%x\n", *(unsigned int *) &mcfg[28]);
  printf ("Creator Revisoin: 0x%x\n", *(unsigned int *) &mcfg[32]);
  for (entry = &mcfg[44]; entry < mcfg + l_mcfg
   ; entry += 16) {
   table = entry; 
   printf ("Lower Base Address Part: 0x%x\n", *(unsigned int *) &table[0]);
   printf ("Upper Base Address Part: 0x%x\n", *(unsigned int *) &table[4]);
   printf ("PCI Offset Number: 0x%hx\n", *(unsigned short *) &table[8]);
   printf ("Start PCI Bus Number: 0x%hhx\n", table[10]);
   printf ("End PCI Bus Number: 0x%hhx\n", table[11]);
  }
 } else {
  printf ("MCFG invalid\n");
 }
}

void scanPCIe (void *mcfg_in) {
 unsigned char *mcfg = mcfg_in, *entry, *mem, *table;
 unsigned int l_mcfg;

 if (checkMCFG (mcfg)) {
 l_mcfg = *(unsigned int *) &mcfg[4];
  for (entry = &mcfg[44]; entry < mcfg + l_mcfg
   ; entry += 16) {
   table = entry; 
   for (mem = *(void **) table
    ; mem < *(unsigned char **) table + 4096 * 32 * 8 * (table[11] - table[10])
    ; mem += 4096) {
    if ((void *) 0xffffffff != *(void **) mem) {
      printf ("Vendor ID: 0x%hx, Device ID: 0x%hx at Address: 0x%p\n"
       , *(unsigned short *) mem, *(unsigned short *) &mem[2], mem);
    }
   }
  }
 } else {
  printf ("Invalid MCFG given\n");
 }
}

int main (int argc, char **argv) {
 void *rsdp, *rsdt, *mcfg;

 printf ("PCI Express Scan 1.0 Alpha\n");
 printf ("Copyright (R) 2019  Asterix\n");
 printf ("Released to the Public Domain\n\n");

 if (1 != argc) {
  printf ("Usage: %s\n", argv[0]);
  return 1;
 }

 rsdp = getRSDP ();
 printRSDP (rsdp);
 printf ("\n");
 rsdt = getRSDT (rsdp);
 printRSDT (rsdt);
 printf ("\n");
 mcfg = getMCFG (rsdt);
 printMCFG (mcfg);
 printf ("\n");
 scanPCIe (mcfg);

 return 0;
}
